;;; -*- Mode: Lisp; Package: KERNEL; Log: code.log -*-
;;;
;;; **********************************************************************
;;; This code was written as part of the CMU Common Lisp project at
;;; Carnegie Mellon University, and has been placed in the public domain.
;;;
(ext:file-comment
  "$Header: /home/cmucl/cvs/src/code/irrat.lisp,v 1.29.2.10 1999/11/19 15:28:42 dtc Exp $")
;;;
;;; **********************************************************************
;;;
;;; This file contains all the irrational functions.
;;;
;;; Author: William Lott.
;;; Long-float support: Douglas Crosher and Raymond Toy, 1998.
;;; 

(in-package "KERNEL")


;;;; Random constants, utility functions, and macros.

(defconstant pi 3.14159265358979323846264338327950288419716939937511L0)
;(defconstant e 2.71828182845904523536028747135266249775724709369996L0)

;;; Make these INLINE, since the call to C is at least as compact as a Lisp
;;; call, and saves number consing to boot.
;;;
(defmacro def-math-rtn (name num-args)
  (let ((function (intern (concatenate 'simple-string
				       "%"
				       (string-upcase name)))))
    `(progn
       (proclaim '(inline ,function))
       (export ',function)
       (alien:def-alien-routine (,name ,function) double-float
	 ,@(let ((results nil))
	     (dotimes (i num-args (nreverse results))
	       (push (list (intern (format nil "ARG-~D" i))
			   'double-float)
		     results)))))))

(eval-when (compile load eval)

#-long-float
(defun handle-reals (function var)
  `((((foreach fixnum single-float bignum ratio))
     (coerce (,function (coerce ,var 'double-float)) 'single-float))
    ((double-float)
     (,function ,var))))

#+long-float
(defun handle-reals (function var)
  `((((foreach fixnum single-float bignum ratio))
     (coerce (,function (coerce ,var 'long-float)) 'single-float))
    ((double-float)
     (coerce (,function (coerce ,var 'long-float)) 'double-float))
    ((long-float)
     (,function ,var))))

); eval-when (compile load eval)


;;;; Stubs for the Unix math library.

;;; Please refer to the Unix man pages for details about these routines.

;;; Trigonometric.
#-x86
(progn
  (def-math-rtn "sin" 1)
  (def-math-rtn "cos" 1)
  (def-math-rtn "tan" 1)
  (def-math-rtn "asin" 1)
  (def-math-rtn "acos" 1)
  (def-math-rtn "atan" 1)
  (def-math-rtn "atan2" 2))

;;; May want to use C library calls for these if available.
#-x86
(progn
  (def-math-rtn "sinh" 1)
  (def-math-rtn "cosh" 1)
  (def-math-rtn "tanh" 1)
  (def-math-rtn "asinh" 1)
  (def-math-rtn "acosh" 1)
  (def-math-rtn "atanh" 1))
  
;;; Exponential and Logarithmic.
#-x86
(progn
  (def-math-rtn "exp" 1)
  (def-math-rtn "log" 1)
  (def-math-rtn "log10" 1)
  (def-math-rtn "pow" 2)
  (def-math-rtn "hypot" 2)
  (def-math-rtn "log1p" 1))
#-(or x86 sparc-v7 sparc-v8 sparc-v9) (def-math-rtn "sqrt" 1)

;;; These are needed for use by byte-compiled files.
#+x86
(progn
  (defun %sin (x)
    (declare (long-float x)
	     (values long-float))
    (%sin x))
  (defun %sin-quick (x)
    (declare (long-float x)
	     (values long-float))
    (%sin-quick x))
  (defun %cos (x)
    (declare (long-float x)
	     (values long-float))
    (%cos x))
  (defun %cos-quick (x)
    (declare (long-float x)
	     (values long-float))
    (%cos-quick x))
  (defun %tan (x)
    (declare (long-float x)
	     (values long-float))
    (%tan x))
  (defun %tan-quick (x)
    (declare (long-float x)
	     (values long-float))
    (%tan-quick x))
  (defun %asin (x)
    (declare (long-float x)
	     (values long-float))
    (atan x (ext:truly-the long-float (sqrt (- 1L0 (* x x))))))
  (defun %acos (x)
    (declare (long-float x)
	     (values long-float))
    (atan (ext:truly-the long-float (sqrt (- 1L0 (* x x)))) x))
  (defun %atan (x)
    (declare (long-float x)
	     (values long-float))
    (%atan x))
  (defun %atan2 (x y)
    (declare (long-float x y)
	     (values long-float))
    (%atan2 x y))
  (defun %exp (x)
    (declare (long-float x)
	     (values long-float))
    (%exp x))
  (defun %expm1 (x)
    (declare (long-float x)
	     (values long-float))
    (%expm1 x))
  (defun %log (x)
    (declare (long-float x)
	     (values long-float))
    (%log x))
  (defun %log10 (x)
    (declare (long-float x)
	     (values long-float))
    (%log10 x))
  (defun %pow (x y)
    (declare (type (long-float 0L0) x)
	     (long-float y)
	     (values (long-float 0L0)))
    (%pow x y))
  (defun %sqrt (x)
    (declare (long-float x)
	     (values long-float))
    (%sqrt x))
  (defun %scalbn (f ex)
    (declare (long-float f)
	     (type (signed-byte 32) ex)
	     (values long-float))
    (%scalbn f ex))
  (defun %scalb (f ex)
    (declare (long-float f ex)
	     (values long-float))
    (%scalb f ex))
  (defun %logb (x)
    (declare (long-float x)
	     (values long-float))
    (%logb x))
  (defun %log1p (x)
    (declare (long-float x)
	     (values long-float))
    (%log1p x))
  ) ; progn



#+(and long-float x86)
(progn
;;; Long-float functions for the X86 port.

(declaim (maybe-inline %hypot %sinh %cosh %tanh %asinh %acosh %atanh))

;;; Quick implementation.
#+nil
(defun %hypot (x y)
  (declare (long-float x y)
	   (values long-float))
  (%sqrt (+ (* x x) (* y y ))))

;;; Avoid overflow and underflow, catch infinite and NaN arguments.
(defun %hypot (x y)
  (declare (long-float x y)
	   (values long-float))
  (cond ((or (float-infinity-p x) (float-infinity-p y))
	 ext:long-float-positive-infinity)
	((float-nan-p x) x)
	((float-nan-p y) y)
	((zerop x) y)
	((zerop y) x)
	(t
	 (let* ((scale (max (kernel:%logb x) (kernel:%logb y)))
		(neg-scale (- scale))
		(x (kernel:%scalb x neg-scale))
		(y (kernel:%scalb y neg-scale)))
	   (kernel:%scalb (%sqrt (+ (* x x) (* y y))) scale)))))

;;; Greater accuracy.
#+nil
(defun %hypot (x y)
  (declare (long-float x y)
	   (values long-float)
	   (optimize (speed 3)))
  (when (or (float-infinity-p x) (float-infinity-p x))
    (return-from %hypot ext:long-float-positive-infinity))
  (when (float-nan-p x)
    (return-from %hypot x))
  (when (float-nan-p y)
    (return-from %hypot y))
  (let ((x (abs x))
	(y (abs y)))
    (when (> y x)
      (let ((temp x))
	(declare (long-float temp))
	(setq x y)
	(setq y temp)))
    (when (zerop x)
      (return-from hypot 0L0))
    (when (zerop y)
      (return-from %hypot x))
    (let ((exp (kernel:%logb x)))
      (when (> (- exp (kernel:%logb y)) 40L0)
	;; raise inexact flag and return |x|
	(return-from %hypot x))
      ;; start computing sqrt(x^2 + y^2)
      (let ((r (- x y)))
	(declare (long-float r))
	(cond ((> r y)	; x/y > 2
	       (setq r (/ x y))
	       (setq r (+ r (ext:truly-the long-float
					   (sqrt (+ 1L0 (* r r)))))))
	      (t	; 1 <= x/y <= 2
	       (setq r (/ r y))
	       (let ((tmp (* r (+ r 2L0))))
		 (setq r (+ (+ (sqrt 2L0) 1L0) r
			    (/ tmp (+ (sqrt 2L0)
				      (ext:truly-the long-float
						     (sqrt (+ 2L0 tmp))))))))))
	(+ x (/ y r))))))


(defun %sinh (x)
  (declare (long-float x))
  (let* ((float-exp-bits (long-float-exp-bits x))
	 (exp-bits (ldb vm:long-float-exponent-byte float-exp-bits)))
    (if (>= exp-bits vm:long-float-normal-exponent-max)	; INF of NaN
	(+ x x)
	(let ((h (if (< float-exp-bits 0) -0.5L0 0.5L0))
	      (x-abs (abs x)))
	  (cond ((< x-abs 23L0)
		 ;; sign(x)*0.5*(E + E/(E+1)), E=expm1(|x|)
		 (let ((d (%expm1 x-abs)))
		   (* h (+ d (/ d (+ d 1L0))))))
		((< x-abs (log most-positive-long-float))
		 (* h (exp x-abs)))
		(t
		 (let ((w (exp (* 0.5L0 x-abs))))
		   (* (* h w) w))))))))

(defun %cosh (x)
  (declare (long-float x))
  (let* ((float-exp-bits (long-float-exp-bits x))
	 (exp-bits (ldb vm:long-float-exponent-byte float-exp-bits)))
    (if (>= exp-bits vm:long-float-normal-exponent-max)	; INF of NaN
	(* x x)
	(let ((x-abs (abs x)))
	  (cond ((< x-abs (* 0.5L0 (log 2L0)))
		 ;; 1+expm1(|x|)^2/(2*exp(|x|))
		 (let* ((d (%expm1 x-abs))
			(w (+ 1L0 d)))
		   (+ 1L0 (/ (* d d) (+ w w)))))
		((< x-abs 23L0)
		 ;; (exp(|x|)+1/exp(|x|)/2
		 (let ((d (exp x-abs)))
		   (+ (* 0.5L0 d) (/ 0.5L0 d))))
		((< x-abs (log most-positive-long-float))
		 ;; 0.5 * exp(|x|)
		 (* 0.5L0 (exp x-abs)))
		(t
		 (let ((d (exp (* 0.5L0 x-abs))))
		   (* (* 0.5L0 d) d))))))))

(defun %tanh (x)
  (declare (long-float x))
  (let* ((float-exp-bits (long-float-exp-bits x))
	 (exp-bits (ldb vm:long-float-exponent-byte float-exp-bits)))
    (if (>= exp-bits vm:long-float-normal-exponent-max)	; INF of NaN
	(if (>= float-exp-bits 0)
	    (+ (/ x) 1L0)	; tanh(+-inf) = +-1
	    (- (/ x) 1L0))	; tanh(NaN) = NaN
	(let* ((x-abs (abs x))
	       (z (cond ((< x-abs 23)
			 (cond ((< x-abs (expt 2L0 -55L0))
				(return-from %tanh (* x (+ 1L0 x))))
			       ((>= x-abs 1L0)
				(let ((d (%expm1 (* 2L0 x-abs))))
				  (- 1L0 (/ 2L0 (+ d 2L0)))))
			       (t
				(let ((d (%expm1 (* -2L0 x-abs))))
				  (/ (- d) (+ d 2L0))))))
			;; |x| > 23, return +-1
			(t
			 1L0))))
	  (if (minusp float-exp-bits) (- z) z)))))

(defun %asinh (x)
  (declare (long-float x))
  (let* ((float-exp-bits (long-float-exp-bits x))
	 (exp-bits (ldb vm:long-float-exponent-byte float-exp-bits)))
    (if (>= exp-bits vm:long-float-normal-exponent-max)	; INF of NaN
	(+ x x)
	(let* ((x-abs (abs x))
	       (w (cond ((> x-abs (/ (sqrt long-float-epsilon)))
			 ;; sign(x)*(log(x)+ln2)) for large |x|
			 (+ (log x-abs) (log 2L0)))
			((> x-abs 2L0)
			 ;; sign(x)*log(2|x|+1/(|x|+sqrt(x*x+1)))
			 (log (+ x-abs x-abs
				 (/ (+ (sqrt (+ 1L0 (* x-abs x-abs)))
				       x-abs)))))
			(t
			 ;; sign(x)*log1p(|x| + x^2/(1 + sqrt(1+x^2)))
			 (let ((x2 (* x-abs x-abs)))
			   (%log1p (+ x-abs
				      (/ x2 (+ 1L0 (%sqrt (+ 1L0 x2)))))))))))
	  (if (minusp float-exp-bits) (- w) w)))))

(defun %acosh (x)
  (declare (long-float x))
  (let* ((float-exp-bits (long-float-exp-bits x))
	 (exp-bits (ldb vm:long-float-exponent-byte float-exp-bits)))
    (cond ((>= exp-bits vm:long-float-normal-exponent-max)	; INF of NaN
	   (+ x x))
	  ((>= x (sqrt (/ long-float-negative-epsilon)))
	   ;; log(x)+ln2
	   (+ (log x) (log 2L0)))
	  ((> x 2L0)
	   ;; log(2x-1/(sqrt(x*x-1)+x))
	   (log (- (+ x x) (/ (+ x (%sqrt (- (* x x) 1L0)))))))
	  (t
	   ;; log1p(t+sqrt(2.0*t+t*t)); where t=x-1.
	   (let ((tmp (- x 1L0)))
	     (%log1p (+ tmp (%sqrt (+ tmp tmp (* tmp tmp))))))))))

(defun %atanh (x)
  (declare (long-float x)
	   (inline float-nan-p))
  (if (float-nan-p x)
      x
      (let ((x-abs (abs x)))
	(if (= x-abs 1L0)
	    (/ x 0L0)
	    ;; 0.5*log1p(2x+2x*x/(1-x))
	    (let* ((x2 (+ x-abs x-abs))
		   (tmp (* 0.5L0 (%log1p (+ x2 (/ (* x2 x-abs)
						  (- 1L0 x-abs)))))))
	      (if (minusp (long-float-exp-bits x)) (- tmp) tmp))))))

) ; end progn


;;;; Power functions.

(defun exp (number)
  "Return e raised to the power NUMBER."
  (number-dispatch ((number number))
    (handle-reals %exp number)
    ((complex)
     (* (exp (realpart number))
	(cis (imagpart number))))))

;;; INTEXP -- Handle the rational base, integer power case.

(defparameter *intexp-maximum-exponent* 10000)

;;; This function precisely calculates base raised to an integral power.  It
;;; separates the cases by the sign of power, for efficiency reasons, as powers
;;; can be calculated more efficiently if power is a positive integer.  Values
;;; of power are calculated as positive integers, and inverted if negative.
;;;
(defun intexp (base power)
  (when (> (abs power) *intexp-maximum-exponent*)
    (cerror "Continue with calculation."
	    "The absolute value of ~S exceeds ~S."
	    power '*intexp-maximum-exponent* base power))
  (cond ((minusp power)
	 (/ (intexp base (- power))))
	((eql base 2)
	 (ash 1 power))
	(t
	 (do ((nextn (ash power -1) (ash power -1))
	      (total (if (oddp power) base 1)
		     (if (oddp power) (* base total) total)))
	     ((zerop nextn) total)
	   (setq base (* base base))
	   (setq power nextn)))))


;;; EXPT  --  Public
;;;
;;;    If an integer power of a rational, use INTEXP above.  Otherwise, do
;;; floating point stuff.  If both args are real, we try %POW right off,
;;; assuming it will return 0 if the result may be complex.  If so, we call
;;; COMPLEX-POW which directly computes the complex result.  We also separate
;;; the complex-real and real-complex cases from the general complex case.
;;;
#+(and long-float x86)
(defun expt (base power)
  "Returns BASE raised to the POWER."
  (if (zerop power)
      (1+ (* base power))
    (labels (;; Determine if the long float is an integer.
	     ;;  0 - not an integer
	     ;;  1 - an odd int
	     ;;  2 - an even int
	     (isint (iexp hi lo)
	       (declare (type (unsigned-byte 15) iexp)
			(type (unsigned-byte 32) hi lo)
			(optimize (speed 3) (safety 0)
				  (ext:inhibit-warnings 3)))
	       (let ((isint 0))
		 (declare (type fixnum isint))
		 (cond ((>= iexp (+ 64 16383))	; exponent >= 64
			(setq isint 2))
		       ((>= iexp 16383)		; exponent >= 0
			(let ((k (- iexp 16383)))	; unbiased exponent
			  (declare (type (mod 64) k))
			  (cond ((> k 31)
				 (let* ((shift (- 63 k))
					(j (ash lo (- shift)))
					(j2 (ash j shift)))
				   (declare (type (mod 32) shift)
					    (type (unsigned-byte 32) j j2))
				   (when (= j2 lo)
				     (setq isint (- 2 (logand j 1))))))
				((= lo 0)
				 (let* ((shift (- 31 k))
					(j (ash hi (- shift)))
					(j2 (ash j shift)))
				   (declare (type (mod 32) shift)
					    (type (unsigned-byte 32) j j2))
				   (when (= j2 hi)
				     (setq isint (- 2 (logand j 1))))))))))
		 isint))
	     (real-expt (x y rtype)
	       (declare (long-float x y))
	       (let* ((x-hi (kernel:long-float-high-bits x))
		      (x-lo (kernel:long-float-low-bits x))
		      (x-exp (kernel:long-float-exp-bits x))
		      (x-iexp (logand x-exp #x7fff))
		      (y-hi (kernel:long-float-high-bits y))
		      (y-lo (kernel:long-float-low-bits y))
		      (y-exp (kernel:long-float-exp-bits y))
		      (y-iexp (logand y-exp #x7fff)))
		 (declare (type (signed-byte 16) x-exp y-exp)
			  (type (unsigned-byte 15) x-iexp y-iexp)
			  (type (unsigned-byte 32) x-hi y-hi x-lo y-lo)
			  (optimize (speed 3) (safety 0)
				    (ext:inhibit-warnings 3)))
		 ;; y==zero: x**0 = 1
		 (when (zerop (logior y-iexp y-hi y-lo))
		   (return-from real-expt (coerce 1L0 rtype)))
		 ;; +-NaN return x+y
		 (when (or (and (= x-iexp #x7fff)
				(or (/= x-hi #x80000000) (/= x-lo 0)))
			   (and (= y-iexp #x7fff)
				(or (/= y-hi #x80000000) (/= y-lo 0))))
		   (return-from real-expt (coerce (+ x y) rtype)))
		 (let ((yisint (if (< x-exp 0) (isint y-iexp y-hi y-lo) 0)))
		   (declare (type (mod 3) yisint))
		   (let ((abs-x (abs x)))
		     (declare (long-float abs-x))
		     ;; special value of y
		     (when (and (zerop y-lo) (= y-hi #x80000000)
				(= y-iexp #x7fff))
		       ;; y is +-inf
		       (return-from real-expt
			 (cond ((= abs-x 1L0)
				;; +-1**inf is NaN
				(coerce (- y y) rtype))
			       ((>= abs-x 1L0)
				;; (|x|>1)**+-inf = inf,0
				(if (>= y-exp 0)
				    (coerce y rtype)
				    (coerce 0L0 rtype)))
			       (t
				;; (|x|<1)**-,+inf = inf,0
				(if (< y-exp 0)
				    (coerce (- y) rtype)
				    (coerce 0L0 rtype))))))
		     ;; special value of x
		     (when (and (zerop x-lo)
				(or (and (= x-hi #x80000000)
					 (or (= x-iexp #x7fff)
					     (= x-iexp 16383)))
				    (and (zerop x-hi) (zerop x-iexp))))
		       ;; x is +-inf,+-1,+-0
		       (let ((z (if (< y-exp 0)
				    (/ 1L0 abs-x)	; z = (1/|x|)
				    abs-x)))
			 (declare (long-float z))
			 (when (< x-exp 0)
			   (cond ((and (= x -1L0) (zerop yisint))
				  ;; (-1)**non-int
				  (let ((y*pi (* y pi)))
				    (declare (long-float y*pi))
				    (return-from real-expt
				      (complex
				       (coerce (%cos y*pi) rtype)
				       (coerce (%sin y*pi) rtype)))))
				 ((= yisint 1)
				  ;; (x<0)**odd = -(|x|**odd)
				  (setq z (- z)))))
			 (return-from real-expt (coerce z rtype))))
		     (if (>= x-exp 0)
			 ;; x>0
			 (coerce (kernel::%pow x y) rtype)
			 ;; x<0
			 (let ((pow (kernel::%pow abs-x y)))
			   (declare (long-float pow))
			   (case yisint
			     (1 ; Odd
			      (coerce (* -1L0 pow) rtype))
			     (2 ; Even
			      (coerce pow rtype))
			     (t ; Non-integer
			      (let ((y*pi (* y pi)))
				(declare (long-float y*pi))
				(complex
				 (coerce (* pow (%cos y*pi)) rtype)
				 (coerce (* pow (%sin y*pi)) rtype)))))))))))
	     ;; Need three variations to handle the different possible
	     ;; result types.
	     (long-expt (x y)
	       (declare (long-float x y))
	       (real-expt x y 'long-float))
	     (double-expt (x y)
	       (declare (long-float x y))
	       (real-expt x y 'double-float))
	     (single-expt (x y)
	       (declare (long-float x y))
	       (real-expt x y 'single-float)))
      (declare (inline real-expt))
      (number-dispatch ((base number) (power number))
        (((foreach fixnum (or bignum ratio) (complex rational)) integer)
	 (intexp base power))
	((single-float rational)
	 (single-expt (coerce base 'long-float) (coerce power 'long-float)))
	((double-float rational)
	 (double-expt (coerce base 'long-float) (coerce power 'long-float)))
	((long-float rational)
	 (long-expt base (coerce power 'long-float)))
	(((foreach fixnum (or bignum ratio) single-float)
	  (foreach ratio single-float))
	 (single-expt (coerce base 'long-float) (coerce power 'long-float)))
	(((foreach fixnum (or bignum ratio) single-float double-float
		   long-float)
	  long-float)
	 (long-expt (coerce base 'long-float) power))
	(((foreach fixnum (or bignum ratio) single-float double-float)
	  double-float)
	 (double-expt (coerce base 'long-float) (coerce power 'long-float)))
	((long-float (foreach double-float single-float))
	 (long-expt base (coerce power 'long-float)))
	((double-float single-float)
	 (double-expt (coerce base 'long-float) (coerce power 'long-float)))
	(((foreach (complex rational) (complex float)) rational)
	 (* (expt (abs base) power)
	    (cis (* power (phase base)))))
	(((foreach fixnum (or bignum ratio) single-float double-float
		   long-float)
	  complex)
	 (if (and (zerop base) (plusp (realpart power)))
	     (* base power)
	     (exp (* power (log base)))))
	(((foreach (complex float) (complex rational))
	  (foreach complex long-float double-float single-float))
	 (if (and (zerop base) (plusp (realpart power)))
	     (* base power)
	     (exp (* power (log base)))))))))

(defun log (number &optional (base nil base-p))
  "Return the logarithm of NUMBER in the base BASE, which defaults to e."
  (if base-p
      (if (zerop base)
	  base				; ANSI spec
	  (/ (log number) (log base)))
      (number-dispatch ((number number))
	(((foreach fixnum bignum ratio))
	 (if (minusp number)
	     (complex (log (- number)) (coerce pi 'single-float))
	     (coerce (%log (coerce number 'long-float)) 'single-float)))
	(((foreach single-float double-float #+long-float long-float))
	 ;; Is (log -0) -infinity (libm.a) or -infinity + i*pi (Kahan)?
	 ;; Since this doesn't seem to be an implementation issue
	 ;; I (pw) take the Kahan result.
	 (if (< (float-sign number) (coerce 0 '(dispatch-type number)))
	     (complex (log (- number)) (coerce pi '(dispatch-type number)))
	     (coerce (%log (coerce number 'long-float))
		     '(dispatch-type number))))
	((complex)
	 (complex-log number)))))

(defun sqrt (number)
  "Return the square root of NUMBER."
  (number-dispatch ((number number))
    (((foreach fixnum bignum ratio))
     (if (minusp number)
	 (complex-sqrt number)
	 (coerce (%sqrt (coerce number 'long-float)) 'single-float)))
    (((foreach single-float double-float #+long-float long-float))
     (if (minusp number)
	 (complex-sqrt number)
	 (coerce (%sqrt (coerce number 'long-float))
		 '(dispatch-type number))))
    ((complex)
     (complex-sqrt number))))


;;;; Trigonometic and Related Functions

(defun abs (number)
  "Returns the absolute value of the number."
  (number-dispatch ((number number))
    (((foreach single-float double-float #+long-float long-float
	       fixnum rational))
     (abs number))
    ((complex)
     (let ((rx (realpart number))
	   (ix (imagpart number)))
       (etypecase rx
	 (rational
	  (sqrt (+ (* rx rx) (* ix ix))))
	 (single-float
	  (coerce (%hypot (coerce rx 'long-float) (coerce ix 'long-float))
		  'single-float))
	 (double-float
	  (coerce (%hypot (coerce rx 'long-float) (coerce ix 'long-float))
		  'double-float))
	 #+long-float
	 (long-float
	  (coerce (%hypot rx ix) 'long-float)))))))

(defun phase (number)
  "Returns the angle part of the polar representation of a complex number.
  For complex numbers, this is (atan (imagpart number) (realpart number)).
  For non-complex positive numbers, this is 0.  For non-complex negative
  numbers this is PI."
  (etypecase number
    (rational
     (if (minusp number)
	 (coerce pi 'single-float)
	 0.0f0))
    (single-float
     (if (minusp (float-sign number))
	 (coerce pi 'single-float)
	 0.0f0))
    (double-float
     (if (minusp (float-sign number))
	 (coerce pi 'double-float)
	 0.0d0))
    #+long-float
    (long-float
     (if (minusp (float-sign number))
	 pi
	 0.0L0))
    (complex
     (atan (imagpart number) (realpart number)))))


(defun sin (number)  
  "Return the sine of NUMBER."
  (number-dispatch ((number number))
    (handle-reals %sin number)
    ((complex)
     (let ((x (realpart number))
	   (y (imagpart number)))
       (complex (* (sin x) (cosh y))
		(* (cos x) (sinh y)))))))

(defun cos (number)
  "Return the cosine of NUMBER."
  (number-dispatch ((number number))
    (handle-reals %cos number)
    ((complex)
     (let ((x (realpart number))
	   (y (imagpart number)))
       (complex (* (cos x) (cosh y))
		(- (* (sin x) (sinh y))))))))

(defun tan (number)
  "Return the tangent of NUMBER."
  (number-dispatch ((number number))
    (handle-reals %tan number)
    ((complex)
     (complex-tan number))))

(defun cis (theta)
  "Return cos(Theta) + i sin(Theta), AKA exp(i Theta)."
  (if (complexp theta)
      (error "Argument to CIS is complex: ~S" theta)
      (complex (cos theta) (sin theta))))

(defun asin (number)
  "Return the arc sine of NUMBER."
  (number-dispatch ((number number))
    ((rational)
     (if (or (> number 1) (< number -1))
	 (complex-asin number)
	 (coerce (%asin (coerce number 'long-float)) 'single-float)))
    (((foreach single-float double-float #+long-float long-float))
     (if (or (> number (coerce 1 '(dispatch-type number)))
	     (< number (coerce -1 '(dispatch-type number))))
	 (complex-asin number)
	 (coerce (%asin (coerce number 'long-float)) '(dispatch-type number))))
    ((complex)
     (complex-asin number))))

(defun acos (number)
  "Return the arc cosine of NUMBER."
  (number-dispatch ((number number))
    ((rational)
     (if (or (> number 1) (< number -1))
	 (complex-acos number)
	 (coerce (%acos (coerce number 'long-float)) 'single-float)))
    (((foreach single-float double-float #+long-float long-float))
     (if (or (> number (coerce 1 '(dispatch-type number)))
	     (< number (coerce -1 '(dispatch-type number))))
	 (complex-acos number)
	 (coerce (%acos (coerce number 'long-float)) '(dispatch-type number))))
    ((complex)
     (complex-acos number))))


(defun atan (y &optional (x nil xp))
  "Return the arc tangent of Y if X is omitted or Y/X if X is supplied."
  (if xp
      (flet ((atan2 (y x)
	       (declare (type long-float y x)
			(values long-float))
	       (if (zerop x)
		   (if (zerop y)
		       (if (plusp (float-sign x))
			   y
			   (float-sign y pi))
		       (float-sign y (/ pi 2)))
		   (%atan2 y x))))
	(number-dispatch ((y number) (x number))
	  ((double-float (foreach double-float single-float fixnum bignum
				  ratio))
	   (coerce (atan2 (coerce y 'long-float) (coerce x 'long-float))
		   'double-float))
	  #+long-float
	  ((long-float
	    (foreach long-float double-float single-float fixnum bignum ratio))
	   (atan2 y (coerce x 'long-float)))
	  #+long-float
	  (((foreach double-float single-float fixnum bignum ratio) long-float)
	   (atan2 (coerce y 'long-float) x))
	  (((foreach single-float fixnum bignum ratio) double-float)
	   (coerce (atan2 (coerce y 'long-float) (coerce x 'long-float))
		   'double-float))
	  (((foreach single-float fixnum bignum ratio)
	    (foreach single-float fixnum bignum ratio))
	   (coerce (atan2 (coerce y 'long-float) (coerce x 'long-float))
		   'single-float))))
      (number-dispatch ((y number))
	(handle-reals %atan y)
	((complex)
	 (complex-atan y)))))

;; It seems that everyone has a C version of sinh, cosh, and
;; tanh. Let's use these for reals because the original
;; implementations based on the definitions lose big in round-off
;; error.  These bad definitions also mean that sin and cos for
;; complex numbers can also lose big.

#+nil
(defun sinh (number)
  "Return the hyperbolic sine of NUMBER."
  (/ (- (exp number) (exp (- number))) 2))

(defun sinh (number)
  "Return the hyperbolic sine of NUMBER."
  (number-dispatch ((number number))
    (handle-reals %sinh number)
    ((complex)
     (let ((x (realpart number))
	   (y (imagpart number)))
       (complex (* (sinh x) (cos y))
		(* (cosh x) (sin y)))))))

#+nil
(defun cosh (number)
  "Return the hyperbolic cosine of NUMBER."
  (/ (+ (exp number) (exp (- number))) 2))

(defun cosh (number)
  "Return the hyperbolic cosine of NUMBER."
  (number-dispatch ((number number))
    (handle-reals %cosh number)
    ((complex)
     (let ((x (realpart number))
	   (y (imagpart number)))
       (complex (* (cosh x) (cos y))
		(* (sinh x) (sin y)))))))

#+nil
(defun tanh (number)
  "Return the hyperbolic tangent of NUMBER."
  (/ (- (exp number) (exp (- number)))
     (+ (exp number) (exp (- number)))))

(defun tanh (number)
  "Return the hyperbolic tangent of NUMBER."
  (number-dispatch ((number number))
    (handle-reals %tanh number)
    ((complex)
     (complex-tanh number))))

#+nil
(defun asinh (number)
  "Return the hyperbolic arc sine of NUMBER."
  (log (+ number (sqrt (1+ (* number number))))))

(defun asinh (number)
  "Return the hyperbolic arc sine of NUMBER."
  (number-dispatch ((number number))
    (handle-reals %asinh number)
    ((complex)
     (complex-asinh number))))

#+nil
(defun acosh (number)
  "Return the hyperbolic arc cosine of NUMBER."
  (* 2 (log (+ (sqrt (/ (1+ number) 2)) (sqrt (/ (1- number) 2))))))

(defun acosh (number)
  "Return the hyperbolic arc cosine of NUMBER."
  (number-dispatch ((number number))
    ((rational)
     ;; acosh is complex if number < 1
     (if (< number 1)
	 (complex-acosh number)
	 (coerce (%acosh (coerce number 'long-float)) 'single-float)))
    (((foreach single-float double-float #+long-float long-float))
     (if (< number (coerce 1 '(dispatch-type number)))
	 (complex-acosh number)
	 (coerce (%acosh (coerce number 'long-float))
		 '(dispatch-type number))))
    ((complex)
     (complex-acosh number))))

#+nil
(defun atanh (number)
  "Return the hyperbolic arc tangent of NUMBER."
  (/ (- (log (1+ number)) (log (- 1 number))) 2))

(defun atanh (number)
  "Return the hyperbolic arc tangent of NUMBER."
  (number-dispatch ((number number))
    ((rational)
     ;; atanh is complex if |number| > 1
     (if (or (> number 1) (< number -1))
	 (complex-atanh number)
	 (coerce (%atanh (coerce number 'long-float)) 'single-float)))
    (((foreach single-float double-float #+long-float long-float))
     (if (or (> number (coerce 1 '(dispatch-type number)))
	     (< number (coerce -1 '(dispatch-type number))))
	 (complex-atanh number)
	 (coerce (%atanh (coerce number 'long-float))
		 '(dispatch-type number))))
    ((complex)
     (complex-atanh number))))


#+old-elfun
(progn
;;; Here are the old definitions of the special functions, for
;;; complex-valued arguments.  Some of these functions suffer from
;;; severe round-off error or unnecessary overflow.
  
(proclaim '(inline mult-by-i))
(defun mult-by-i (number)
  (complex (- (imagpart number))
	   (realpart number)))

(defun complex-sqrt (x)
  (exp (/ (log x) 2)))

(defun complex-log (x)
  (complex (log (abs x))
	   (phase x)))

(defun complex-atanh (number)
  (/ (- (log (1+ number)) (log (- 1 number))) 2))

(defun complex-tanh (number)
  (/ (- (exp number) (exp (- number)))
     (+ (exp number) (exp (- number)))))

(defun complex-acos (number)
  (* -2 (mult-by-i (log (+ (sqrt (/ (1+ number) 2))
			   (mult-by-i (sqrt (/ (- 1 number) 2))))))))

(defun complex-acosh (number)
  (* 2 (log (+ (sqrt (/ (1+ number) 2)) (sqrt (/ (1- number) 2))))))

(defun complex-asin (number)
  (- (mult-by-i (log (+ (mult-by-i number) (sqrt (- 1 (* number number))))))))

(defun complex-asinh (number)
  (log (+ number (sqrt (1+ (* number number))))))

(defun complex-atan (y)
  (let ((im (imagpart y))
	(re (realpart y)))
    (/ (- (log (complex (- 1 im) re))
	  (log (complex (+ 1 im) (- re))))
       (complex 0 2))))

(defun complex-tan (number)
  (let* ((num (sin number))
	 (denom (cos number)))
    (if (zerop denom) (error "~S undefined tangent." number)
	(/ num denom))))
)

#-old-specfun
(progn
;;;;
;;;; This is a set of routines that implement many elementary
;;;; transcendental functions as specified by ANSI Common Lisp.  The
;;;; implementation is based on Kahan's paper.
;;;;
;;;; I believe I have accurately implemented the routines and are
;;;; correct, but you may want to check for your self.
;;;;
;;;; These functions are written for CMU Lisp and take advantage of
;;;; some of the features available there.  It may be possible,
;;;; however, to port this to other Lisps.
;;;;
;;;; Some functions are significantly more accurate than the original
;;;; definitions in CMU Lisp.  In fact, some functions in CMU Lisp
;;;; give the wrong answer like (acos #c(-2.0 0.0)), where the true
;;;; answer is pi + i*log(2-sqrt(3)).
;;;;
;;;; All of the implemented functions will take any number for an
;;;; input, but the result will always be a either a complex
;;;; single-float or a complex double-float.
;;;;
;;;; General functions
;;;;   complex-sqrt
;;;;   complex-log
;;;;   complex-atanh
;;;;   complex-tanh
;;;;   complex-acos
;;;;   complex-acosh
;;;;   complex-asin
;;;;   complex-asinh
;;;;   complex-atan
;;;;   complex-tan
;;;;
;;;; Utility functions:
;;;;   scalb logb
;;;;
;;;; Internal functions:
;;;;    square coerce-to-complex-type cssqs complex-log-scaled
;;;;
;;;;
;;;; Please send any bug reports, comments, or improvements to Raymond
;;;; Toy at toy@rtp.ericsson.se.
;;;;
;;;; References
;;;;
;;;; Kahan, W. "Branch Cuts for Complex Elementary Functions, or Much
;;;; Ado About Nothing's Sign Bit" in Iserles and Powell (eds.) "The
;;;; State of the Art in Numerical Analysis", pp. 165-211, Clarendon
;;;; Press, 1987
;;;;
(declaim (inline square))
(declaim (ftype (function (long-float) (long-float 0L0)) square))
(defun square (x)
  (declare (long-float x)
	   (values (long-float 0L0)))
  (* x x))

;; If you have these functions in libm, perhaps they should be used
;; instead of these Lisp versions.  These versions are probably good
;; enough, especially since they are portable.

(declaim (inline scalb))
(defun scalb (x n)
  "Compute 2^N * X without compute 2^N first (use properties of the
underlying floating-point format"
  (declare (type long-float x)
	   (type long-float-exponent n))
  (scale-float x n))

(defun logb (x)
  "Compute an integer N such that 1 <= |2^N * x| < 2.
For the special cases, the following values are used:

    x             logb
   NaN            NaN
   +/- infinity   +infinity
   0              -infinity
"
  (declare (type long-float x))
  (cond ((float-nan-p x)
	 x)
	((float-infinity-p x)
	 #.ext:long-float-positive-infinity)
	((zerop x)
	 ;; The answer is negative infinity, but we are supposed to
	 ;; signal divide-by-zero.
	 ;; (error 'division-by-zero :operation 'logb :operands (list x))
	 (/ -1.0L0 x)
	 )
	(t
	 (multiple-value-bind (signif expon sign)
	     (decode-float x)
	   (declare (ignore signif sign))
	   ;; decode-float is almost right, except that the exponent
	   ;; is off by one
	   (1- expon)))))

;; This function is used to create a complex number of the appropriate
;; type.

(declaim (inline coerce-to-complex-type))
(defun coerce-to-complex-type (x y z)
  "Create complex number with real part X and imaginary part Y such that
it has the same type as Z.  If Z has type (complex rational), the X
and Y are coerced to single-float."
  (declare (long-float x y)
	   (number z))
  (cond
    #+long-float
    ((subtypep (type-of (realpart z)) 'long-float)
     (complex x y))
    ((subtypep (type-of (realpart z)) 'double-float)
     (complex (float x 1d0) (float y 1d0)))
    (t
     ;; Convert anything that's not a double-float to a single-float.
     (complex (float x 1f0) (float y 1f0)))))

(defun cssqs (z)
  ;; Compute |(x+i*y)/2^k|^2 scaled to avoid over/underflow. The
  ;; result is r + i*k, where k is an integer.
  
  ;; Save all FP flags
  (let ((x (float (realpart z) 1L0))
	(y (float (imagpart z) 1L0))
	(k 0)
	(rho 0L0))
    (declare (long-float x y)
	     (type (long-float 0L0) rho)
	     (fixnum k))
    ;; Would this be better handled using an exception handler to
    ;; catch the overflow or underflow signal?  For now, we turn all
    ;; traps off and look at the accrued exceptions to see if any
    ;; signal would have been raised.
    (with-float-traps-masked (:underflow :overflow)
      (setf rho (+ (square x) (square y)))
      (cond ((and (or (float-nan-p rho)
		      (float-infinity-p rho))
		  (or (float-infinity-p (abs x))
		      (float-infinity-p (abs y))))
	     (setf rho #.ext:long-float-positive-infinity))
	    ((let ((threshold #.(/ least-positive-long-float
				   long-float-epsilon))
		   (traps (ldb vm::float-sticky-bits
			       (vm:floating-point-modes))))
	       ;; Overflow raised or (underflow raised and rho <
	       ;; lambda/eps)
	       (or (not (zerop (logand vm:float-overflow-trap-bit traps)))
		   (and (not (zerop (logand vm:float-underflow-trap-bit traps)))
			(< rho threshold))))
	     (setf k (logb (max (abs x) (abs y))))
	     (setf rho (+ (square (scalb x (- k)))
			  (square (scalb y (- k))))))))
    (values rho k)))

(defun complex-sqrt (z)
  "Principle square root of Z

Z may be any number, but the result is always a complex."
  (declare (number z))
  (multiple-value-bind (rho k)
      (cssqs z)
    (declare (type (long-float 0L0) rho)
	     (fixnum k))
    (let ((x (float (realpart z) 1.0L0))
	  (y (float (imagpart z) 1.0L0))
	  (eta 0L0)
	  (nu 0L0))
      (declare (long-float x y eta nu))

      (if (not (float-nan-p x))
	  (setf rho (+ (scalb (abs x) (- k)) (sqrt rho))))

      (cond ((oddp k)
	     (setf k (ash k -1)))
	    (t
	     (setf k (1- (ash k -1)))
	     (setf rho (+ rho rho))))

      (setf rho (scalb (sqrt rho) k))

      (setf eta rho)
      (setf nu y)

      (when (/= rho 0L0)
	    (when (not (float-infinity-p (abs nu)))
		  (setf nu (/ (/ nu rho) 2L0)))
	    (when (< x 0L0)
		  (setf eta (abs nu))
		  (setf nu (float-sign y rho))))
      (coerce-to-complex-type eta nu z))))
    
(defun complex-log-scaled (z j)
  "Compute log(2^j*z).

This is for use with J /= 0 only when |z| is huge."
  (declare (number z)
	   (fixnum j))
  ;; The constants t0, t1, t2 should be evaluated to machine
  ;; precision.  In addition, Kahan says the accuracy of log1p
  ;; influences the choices of these constants but doesn't say how to
  ;; choose them.  We'll just assume his choices matches our
  ;; implementation of log1p.
  (let ((t0 #.(/ 1 (sqrt 2.0L0)))
	(t1 1.2L0)
	(t2 3L0)
	(ln2 #.(log 2L0))
	(x (float (realpart z) 1.0L0))
	(y (float (imagpart z) 1.0L0)))
    (multiple-value-bind (rho k)
	(cssqs z)
      (declare (type (long-float 0L0) rho)
	       (fixnum k))
      (let ((beta (max (abs x) (abs y)))
	    (theta (min (abs x) (abs y))))
	(declare (type (long-float 0L0) beta theta))
	(if (and (zerop k) (< t0 beta)
		 (or (<= beta t1) (< rho t2)))
	    (setf rho (/ (%log1p (+ (* (- beta 1.0L0) (+ beta 1.0L0))
				    (* theta theta)))
			 2L0))
	    (setf rho (+ (/ (log rho) 2L0) (* (+ k j) ln2))))
	(setf theta (atan y x))
	(coerce-to-complex-type rho theta z)))))

(defun complex-log (z)
  "Log of Z = log |Z| + i * arg Z

Z may be any number, but the result is always a complex."
  (declare (number z))
  (complex-log-scaled z 0))
	       
;; Let us note the following "strange" behavior.  atanh 1.0 is
;; +infinity, but the following code returns approx 176 + i*pi/4. The
;; reason for the imaginary part is caused by the fact that arg i*y is
;; never 0 since we have positive and negative zeroes.

(defun complex-atanh (z)
  "Compute atanh z = (log(1+z) - log(1-z))/2"
  (declare (number z))
  (let* (;; Constants
	 (theta #.(/ (sqrt most-positive-long-float) 4.0L0))
	 (rho #.(/ 4.0L0 (sqrt most-positive-long-float)))
	 (half-pi #.(/ pi 2.0L0))
	 (rp (float (realpart z) 1.0L0))
	 (beta (float-sign rp 1.0L0))
	 (x (* beta rp))
	 (y (* beta (- (float (imagpart z) 1.0L0))))
	 (eta 0.0L0)
	 (nu 0.0L0))
    (declare (long-float theta rho half-pi rp beta y eta nu)
	     (type (long-float 0L0) x))
    (cond ((or (> x theta)
	       (> (abs y) theta))
	   ;; To avoid overflow...
	   (setf eta (float-sign y half-pi))
	   ;; nu is real part of 1/(x + iy).  This is x/(x^2+y^2),
	   ;; which can cause overflow.  Arrange this computation so
	   ;; that it won't overflow.
	   (setf nu (let* ((x-bigger (> x (abs y)))
			   (r (if x-bigger (/ y x) (/ x y)))
			   (d (+ 1.0L0 (* r r))))
		      (declare (long-float r d))
		      (if x-bigger
			  (/ (/ x) d)
			  (/ (/ r y) d)))))
	  ((= x 1.0L0)
	   ;; Should this be changed so that if y is zero, eta is set
	   ;; to +infinity instead of approx 176?  In any case
	   ;; tanh(176) is 1.0L0 within working precision.
	   (let ((t1 (+ 4L0 (square y)))
		 (t2 (+ (abs y) rho)))
	     (declare (type (long-float 0L0) t1 t2))
	     #+nil
	     (setf eta (log (/ (sqrt (sqrt t1))) (sqrt t2)))
	     (setf eta (* 0.5L0 (log (the (long-float 0.0L0)
					  (/ (sqrt t1) t2)))))
	     (setf nu (* 0.5L0
			 (float-sign y (+ half-pi (%atan (* 0.5L0 t2))))))))
	  (t
	   (let ((t1 (+ (abs y) rho)))
	     (declare (long-float t1))
	     ;; Normal case using log1p(x) = log(1 + x)
	     (setf eta (* 0.25L0 (%log1p (/ (* 4.0L0 x) (+ (square (- 1.0L0 x))
							   (square t1))))))
	     (setf nu (* 0.5L0 (atan (* 2.0L0 y) (- (* (- 1.0L0 x) (+ 1.0L0 x))
						    (square t1))))))))
    (coerce-to-complex-type (* beta eta) (- (* beta nu)) z)))

(defun complex-tanh (z)
  "Compute tanh z = sinh z / cosh z"
  (declare (number z))
  (let ((x (float (realpart z) 1.0L0))
	(y (float (imagpart z) 1.0L0)))
    (declare (long-float x y))
    (cond ((> (abs x)
	      #+(or x86 hpux) #.(/ (+ (log 2.0L0)
				      (log most-positive-long-float))
				   4L0))
	   (complex (float-sign x) (float-sign y 0.0L0)))
	  (t
	   (let* ((tv (%tan y))
		  (beta (+ 1.0L0 (* tv tv)))
		  (s (sinh x))
		  (rho (sqrt (+ 1.0L0 (* s s)))))
	     (declare (long-float tv s)
		      (type (long-float 0.0L0) beta rho))
	     (if (float-infinity-p (abs tv))
		 (coerce-to-complex-type (/ rho s) (/ tv) z)
		 (let ((den (+ 1.0L0 (* beta s s))))
		   (coerce-to-complex-type (/ (* beta rho s) den)
					   (/ tv den)
					   z))))))))

;; Kahan says we should only compute the parts needed.  Thus, the
;; realpart's below should only compute the real part, not the whole
;; complex expression.  Doing this can be important because we may get
;; spurious signals that occur in the part that we are not using.
;;
;; However, we take a pragmatic approach and just use the whole
;; expression.

;; NOTE: The formula given by Kahan is somewhat ambiguous in whether
;; it's the conjugate of the square root or the square root of the
;; conjugate.  This needs to be checked.

;; I checked.  It doesn't matter because (conjugate (sqrt z)) is the
;; same as (sqrt (conjugate z)) for all z.  This follows because
;;
;; (conjugate (sqrt z)) = exp(0.5*log |z|)*exp(-0.5*j*arg z).
;;
;; (sqrt (conjugate z)) = exp(0.5*log|z|)*exp(0.5*j*arg conj z)
;;
;;.and these two expressions are equal if and only if arg conj z =
;;-arg z, which is clearly true for all z.

(defun complex-acos (z)
  "Compute acos z = pi/2 - asin z

Z may be any number, but the result is always a complex."
  (declare (number z))
  (let ((sqrt-1+z (complex-sqrt (+ 1 z)))
	(sqrt-1-z (complex-sqrt (- 1 z))))
    (with-float-traps-masked (:divide-by-zero)
      (complex (* 2 (atan (/ (realpart sqrt-1-z)
			     (realpart sqrt-1+z))))
	       (asinh (imagpart (* (conjugate sqrt-1+z)
				   sqrt-1-z)))))))

(defun complex-acosh (z)
  "Compute acosh z = 2 * log(sqrt((z+1)/2) + sqrt((z-1)/2))

Z may be any number, but the result is always a complex."
  (declare (number z))
  (let ((sqrt-z-1 (complex-sqrt (- z 1)))
	(sqrt-z+1 (complex-sqrt (+ z 1))))
    (with-float-traps-masked (:divide-by-zero)
      (complex (asinh (realpart (* (conjugate sqrt-z-1)
				   sqrt-z+1)))
	       (* 2 (atan (/ (imagpart sqrt-z-1)
			     (realpart sqrt-z+1))))))))


(defun complex-asin (z)
  "Compute asin z = asinh(i*z)/i

Z may be any number, but the result is always a complex."
  (declare (number z))
  (let ((sqrt-1-z (complex-sqrt (- 1 z)))
	(sqrt-1+z (complex-sqrt (+ 1 z))))
    (with-float-traps-masked (:divide-by-zero)
      (complex (atan (/ (realpart z)
			(realpart (* sqrt-1-z sqrt-1+z))))
	       (asinh (imagpart (* (conjugate sqrt-1-z)
				   sqrt-1+z)))))))

(defun complex-asinh (z)
  "Compute asinh z = log(z + sqrt(1 + z*z))

Z may be any number, but the result is always a complex."
  (declare (number z))
  ;; asinh z = -i * asin (i*z)
  (let* ((iz (complex (- (imagpart z)) (realpart z)))
	 (result (complex-asin iz)))
    (complex (imagpart result)
	     (- (realpart result)))))
	 
(defun complex-atan (z)
  "Compute atan z = atanh (i*z) / i

Z may be any number, but the result is always a complex."
  (declare (number z))
  ;; atan z = -i * atanh (i*z)
  (let* ((iz (complex (- (imagpart z)) (realpart z)))
	 (result (complex-atanh iz)))
    (complex (imagpart result)
	     (- (realpart result)))))

(defun complex-tan (z)
  "Compute tan z = -i * tanh(i * z)

Z may be any number, but the result is always a complex."
  (declare (number z))
  ;; tan z = -i * tanh(i*z)
  (let* ((iz (complex (- (imagpart z)) (realpart z)))
	 (result (complex-tanh iz)))
    (complex (imagpart result)
	     (- (realpart result)))))
)
